@@ -494,11 +494,6 @@ static int diff_list_apply_options(
 
 	/* Don't set GIT_DIFFCAPS_USE_DEV - compile time option in core git */
 
-	/* Don't trust nanoseconds; we do not load nanos from disk */
-#ifdef GIT_USE_NSEC
-	diff->diffcaps = diff->diffcaps | GIT_DIFFCAPS_TRUST_NANOSECS;
-#endif
-
 	/* If not given explicit `opts`, check `diff.xyz` configs */
 	if (!opts) {
 		int context = git_config__get_int_force(cfg, "diff.context", 3);
@@ -699,38 +694,6 @@ int git_diff__oid_for_entry(
 	return error;
 }
 
-static bool diff_time_eq(
-	const git_index_time *a, const git_index_time *b, bool use_nanos)
-{
-	return a->seconds == b->seconds &&
-		(!use_nanos || a->nanoseconds == b->nanoseconds);
-}
-
-/*
- * Test if the given index time is newer than the given existing index entry.
- * If the timestamps are exactly equivalent, then the given index time is
- * considered "racily newer" than the existing index entry.
- */
-static bool diff_newer_than_index(
-	const git_index_time *a, const git_index *b, bool use_nanos)
-{
-	bool is_newer = false;
-
-	if(!b)
-		return false;
-
-	is_newer = is_newer || (a->seconds > (int32_t) b->stamp.mtime.tv_sec);
-	is_newer = is_newer || (!use_nanos &&
-		(a->seconds == (int32_t) b->stamp.mtime.tv_sec));
-	if(use_nanos)
-	{
-		is_newer = is_newer || ((a->seconds == (int32_t) b->stamp.mtime.tv_sec) &&
-			(a->nanoseconds >= (uint32_t) b->stamp.mtime.tv_nsec));
-	}
-
-	return is_newer;
-}
-
 typedef struct {
 	git_repository *repo;
 	git_iterator *old_iter;
@@ -863,13 +826,7 @@ static int maybe_modified(
 	 */
 	} else if (git_oid_iszero(&nitem->id) && new_is_workdir) {
 		bool use_ctime = ((diff->diffcaps & GIT_DIFFCAPS_TRUST_CTIME) != 0);
-#ifdef GIT_USE_NSEC
-		bool use_nanos = ((diff->diffcaps & GIT_DIFFCAPS_TRUST_NANOSECS) != 0);
-#else
-		bool use_nanos = false;
-#endif
-		git_index *index;
-		git_iterator_index(&index, info->new_iter);
+		git_index *index = git_iterator_index(info->new_iter);
 
 		status = GIT_DELTA_UNMODIFIED;
 
@@ -886,13 +843,12 @@ static int maybe_modified(
 			modified_uncertain =
 				(oitem->file_size <= 0 && nitem->file_size > 0);
 		}
-		else if (!diff_time_eq(&oitem->mtime, &nitem->mtime, use_nanos) ||
-			(use_ctime &&
-			 !diff_time_eq(&oitem->ctime, &nitem->ctime, use_nanos)) ||
+		else if (!git_index_time_eq(&oitem->mtime, &nitem->mtime) ||
+			(use_ctime && !git_index_time_eq(&oitem->ctime, &nitem->ctime)) ||
 			oitem->ino != nitem->ino ||
 			oitem->uid != nitem->uid ||
 			oitem->gid != nitem->gid ||
-			diff_newer_than_index(&nitem->mtime, index, use_nanos))
+			git_index_entry_newer_than_index(nitem, index))
 		{
 			status = GIT_DELTA_MODIFIED;
 			modified_uncertain = true;
@@ -1023,15 +979,14 @@ static int iterator_advance_into(
 	return error;
 }
 
-static int iterator_advance_over_with_status(
+static int iterator_advance_over(
 	const git_index_entry **entry,
 	git_iterator_status_t *status,
 	git_iterator *iterator)
 {
-	int error;
+	int error = git_iterator_advance_over(entry, status, iterator);
 
-	if ((error = git_iterator_advance_over_with_status(
-			entry, status, iterator)) == GIT_ITEROVER) {
+	if (error == GIT_ITEROVER) {
 		*entry = NULL;
 		error = 0;
 	}
@@ -1099,7 +1054,7 @@ static int handle_unmatched_new_item(
 				return iterator_advance(&info->nitem, info->new_iter);
 
 			/* iterate into dir looking for an actual untracked file */
-			if ((error = iterator_advance_over_with_status(
+			if ((error = iterator_advance_over(
 					&info->nitem, &untracked_state, info->new_iter)) < 0)
 				return error;
 
@@ -1128,17 +1083,13 @@ static int handle_unmatched_new_item(
 		if (recurse_into_dir) {
 			error = iterator_advance_into(&info->nitem, info->new_iter);
 
-			/* if real error or no error, proceed with iteration */
-			if (error != GIT_ENOTFOUND)
-				return error;
-			giterr_clear();
+			/* if directory is empty, can't advance into it, so skip it */
+			if (error == GIT_ENOTFOUND) {
+				giterr_clear();
+				error = iterator_advance(&info->nitem, info->new_iter);
+			}
 
-			/* if directory is empty, can't advance into it, so either skip
-			 * it or ignore it
-			 */
-			if (contains_oitem)
-				return iterator_advance(&info->nitem, info->new_iter);
-			delta_type = GIT_DELTA_IGNORED;
+			return error;
 		}
 	}
 
@@ -1274,9 +1225,8 @@ int git_diff__from_iterators(
 
 	/* make iterators have matching icase behavior */
 	if (DIFF_FLAG_IS_SET(diff, GIT_DIFF_IGNORE_CASE)) {
-		if ((error = git_iterator_set_ignore_case(old_iter, true)) < 0 ||
-			(error = git_iterator_set_ignore_case(new_iter, true)) < 0)
-			goto cleanup;
+		git_iterator_set_ignore_case(old_iter, true);
+		git_iterator_set_ignore_case(new_iter, true);
 	}
 
 	/* finish initialization */
@@ -1414,7 +1364,7 @@ int git_diff_tree_to_index(
 
 	DIFF_FROM_ITERATORS(
 		git_iterator_for_tree(&a, old_tree, &a_opts), iflag,
-		git_iterator_for_index(&b, index, &b_opts), iflag
+		git_iterator_for_index(&b, repo, index, &b_opts), iflag
 	);
 
 	/* if index is in case-insensitive order, re-sort deltas to match */
@@ -1438,7 +1388,7 @@ int git_diff_index_to_workdir(
 		return error;
 
 	DIFF_FROM_ITERATORS(
-		git_iterator_for_index(&a, index, &a_opts),
+		git_iterator_for_index(&a, repo, index, &a_opts),
 		GIT_ITERATOR_INCLUDE_CONFLICTS,
 
 		git_iterator_for_workdir(&b, repo, index, NULL, &b_opts),
@@ -1515,8 +1465,8 @@ int git_diff_index_to_index(
 	assert(diff && old_index && new_index);
 
 	DIFF_FROM_ITERATORS(
-		git_iterator_for_index(&a, old_index, &a_opts), GIT_ITERATOR_DONT_IGNORE_CASE,
-		git_iterator_for_index(&b, new_index, &b_opts), GIT_ITERATOR_DONT_IGNORE_CASE
+		git_iterator_for_index(&a, repo, old_index, &a_opts), GIT_ITERATOR_DONT_IGNORE_CASE,
+		git_iterator_for_index(&b, repo, new_index, &b_opts), GIT_ITERATOR_DONT_IGNORE_CASE
 	);
 
 	/* if index is in case-insensitive order, re-sort deltas to match */
@@ -1702,6 +1652,7 @@ int git_diff_format_email__append_header_tobuf(
 	const git_oid *id,
 	const git_signature *author,
 	const char *summary,
+	const char *body,
 	size_t patch_no,
 	size_t total_patches,
 	bool exclude_patchno_marker)
@@ -1741,6 +1692,13 @@ int git_diff_format_email__append_header_tobuf(
 
 	error = git_buf_printf(out, "%s\n\n", summary);
 
+	if (body) {
+		git_buf_puts(out, body);
+
+		if (out->ptr[out->size - 1] != '\n')
+			git_buf_putc(out, '\n');
+	}
+
 	return error;
 }
 
@@ -1818,7 +1776,7 @@ int git_diff_format_email(
 
 	error = git_diff_format_email__append_header_tobuf(out,
 				opts->id, opts->author, summary == NULL ? opts->summary : summary,
-				opts->patch_no, opts->total_patches, ignore_marker);
+				opts->body, opts->patch_no, opts->total_patches, ignore_marker);
 
 	if (error < 0)
 		goto on_error;
@@ -1861,6 +1819,7 @@ int git_diff_commit_as_email(
 	opts.total_patches = total_patches;
 	opts.id = git_commit_id(commit);
 	opts.summary = git_commit_summary(commit);
+	opts.body = git_commit_body(commit);
 	opts.author = git_commit_author(commit);
 
 	if ((error = git_diff__commit(&diff, repo, commit, diff_opts)) < 0)
