@@ -1,5 +1,5 @@
 package Git::Raw;
-$Git::Raw::VERSION = '0.57'; # TRIAL
+$Git::Raw::VERSION = '0.59'; # TRIAL
 use strict;
 use warnings;
 
@@ -13,6 +13,8 @@ use Git::Raw::Blob;
 use Git::Raw::Commit;
 use Git::Raw::Reference;
 use Git::Raw::Repository;
+use Git::Raw::Stash;
+use Git::Raw::Stash::Progress;
 use Git::Raw::Tree;
 
 =for HTML
@@ -33,7 +35,7 @@ Git::Raw - Perl bindings to the Git linkable library (libgit2)
 
 =head1 VERSION
 
-version 0.57
+version 0.59
 
 =head1 DESCRIPTION
 
