@@ -1,5 +1,5 @@
 package Git::Raw::Packbuilder;
-$Git::Raw::Packbuilder::VERSION = '0.57'; # TRIAL
+$Git::Raw::Packbuilder::VERSION = '0.59'; # TRIAL
 use strict;
 use warnings;
 use Carp;
@@ -29,7 +29,17 @@ Git::Raw::Packbuilder - Git packbuilder class
 
 =head1 VERSION
 
-version 0.57
+version 0.59
+
+=head1 CONSTANTS
+
+=head2 ADDING_OBJECTS
+
+Objects are being added to the object database.
+
+=head2 DELTAFICATION
+
+Deltas are calculated/resolved.
 
 =head1 AUTHOR
 
