@@ -1,5 +1,5 @@
 package Git::Raw::Repository;
-$Git::Raw::Repository::VERSION = '0.57'; # TRIAL
+$Git::Raw::Repository::VERSION = '0.59'; # TRIAL
 use strict;
 use warnings;
 
@@ -11,7 +11,7 @@ Git::Raw::Repository - Git repository class
 
 =head1 VERSION
 
-version 0.57
+version 0.59
 
 =head1 SYNOPSIS
 
@@ -513,9 +513,9 @@ C<%checkout_opts> values.  Valid fields for C<%merge_opts> are
 
 =over 4
 
-=item * "tree_flags"
+=item * "flags"
 
-Merge tree flags. Valid values include:
+Merge flags. Valid values include:
 
 =over 8
 
@@ -825,6 +825,15 @@ directory. See C<Git::Raw::Repository-E<gt>merge()> for valid C<%merge_opts>
 and C<%checkout_opts> values. For merge commits C<$mainline> specifies the
 parent.
 
+=head2 revparse( $spec )
+
+Parse the revision string C<$spec> to for C<from>, C<to> and C<intent>. Returns a
+list of objects in list context and the number of objects parsed from C<$spec>
+in scalar context.
+
+	my ($from, $to) = $repo -> revparse('HEAD~..HEAD');
+	print "Range is $from -> $to", "\n";
+
 =head2 state( )
 
 Determine the state of the repository. One of the following values is returned:
